package info.textgrid.services.aggregator.epub;

import info.textgrid.namespaces.metadata.core._2010.MetadataContainerType;
import info.textgrid.namespaces.metadata.core._2010.ObjectType;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.AuthFault;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.IoFault;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.MetadataParseFault;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.ObjectNotFoundFault;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.ProtocolNotImplementedFault;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.TGCrudService;
import info.textgrid.services.aggregator.ITextGridRep;
import info.textgrid.services.aggregator.RESTUtils;
import info.textgrid.services.aggregator.TextGridRepProvider;
import info.textgrid.services.aggregator.teicorpus.TEICorpusSerializer;
import info.textgrid.utils.export.filenames.DefaultFilenamePolicy;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.text.MessageFormat;
import java.util.logging.Logger;

import javax.servlet.ServletContext;
import javax.ws.rs.GET;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.Response.Status;
import javax.xml.transform.stream.StreamSource;

import net.sf.saxon.s9api.Processor;
import net.sf.saxon.s9api.SaxonApiException;
import net.sf.saxon.s9api.XsltCompiler;
import net.sf.saxon.s9api.XsltExecutable;
import net.sf.saxon.s9api.XsltTransformer;

import org.apache.cxf.jaxrs.model.wadl.Description;

import com.google.common.io.Files;
import com.google.common.io.InputSupplier;

@Path("/epub")
@Description("Converts the given TEI object or the aggregation of TEI objects to an E-Book in EPUB format")
public class EPUB {

	private static final String EPUB_XSL = "/WEB-INF/stylesheets/epub.xsl";

	private ITextGridRep repository = TextGridRepProvider.getInstance();

	final Logger logger = Logger.getLogger("info.textgrid.services.aggregator.EPUB");

	private XsltExecutable teiToEpub;

	private final Processor xsltProcessor;

	@Context
	private ServletContext servlet;

	private XsltExecutable getTeiToEpub() {
		if (teiToEpub == null) {
			final XsltCompiler xsltCompiler = xsltProcessor.newXsltCompiler();
			try {
				URL stylesheet;
				if (servlet == null) {
					logger.info("No servlet context, trying fallback property");
					final String dir = System.getProperty("webapp.directory");
					if (dir == null)
						throw new IllegalStateException(
								"Could not find stylesheet: Neither ServletContext nor fallback property webapp.directory have been defined.");
					stylesheet = new URL("file://" + dir + EPUB_XSL);
				} else {
					stylesheet = servlet.getResource(EPUB_XSL);
				}
				logger.info("Stylesheet: " + stylesheet.toExternalForm());
				teiToEpub = xsltCompiler.compile(new StreamSource(stylesheet
						.toString()));
			} catch (final MalformedURLException e) {
				throw new IllegalStateException(e);
			} catch (final SaxonApiException e) {
				throw new IllegalStateException(e);
			}
		}
		return teiToEpub;
	}

	public EPUB(final ITextGridRep repository) {
		this.repository = repository;
		xsltProcessor = new Processor(false);
	}

	@GET
	@Path(value = "/{object}")
	@Produces(value = "application/epub+zip")
	public Response get(
			@Description("The TextGrid URI of the object to convert, either a TEI object or an aggregation of TEI (and maybe other) objects")
			@PathParam("object") final URI uri,
			@Description("URL of an alternative stylesheet to use. Must be compatible.") @QueryParam("stylesheet") final URI xsluri,
			@Description("Session ID for accessing protected objects") @QueryParam("sid") final String sid)
					throws ObjectNotFoundFault, MetadataParseFault, IoFault, AuthFault,
					ProtocolNotImplementedFault {
		logger.fine("EPUB called for root object: " + uri);
		final TGCrudService crud = repository.getCRUDService();
		try {
			final MetadataContainerType container = crud.readMetadata(sid,
					null, uri.toString());
			final ObjectType rootObject = container.getObject();
			final String mimeType = rootObject.getGeneric().getProvided()
					.getFormat();
			final boolean aggregation = mimeType.contains("aggregation");
			if (!aggregation && !mimeType.matches("^text/.*xml.*")) {
				final String errorMsg = "The EPUB export can only convert aggregations or XML documents to EPUB, however, the document {0} you referred to has the MIME type {1}.";
				logger.warning("Failing with: " + errorMsg);
				return Response.status(Status.UNSUPPORTED_MEDIA_TYPE)
						.entity(MessageFormat.format(errorMsg, uri, mimeType))
						.type("text/plain").build();
			}

			final File workingDir = Files.createTempDir();
			logger.fine("Using " + workingDir + " to build the E-Book");

			final File corpus = new File(workingDir, "corpus.xml");
			if (aggregation) {
				// First, use the aggregator to create a TEI corpus file to build on
				final TEICorpusSerializer corpusSerializer = new TEICorpusSerializer(
						rootObject, true, sid);
				final FileOutputStream corpusOutput = new FileOutputStream(corpus);
				corpusSerializer.write(corpusOutput);
				corpusOutput.close();
			} else {
				final InputStream tei = repository.getContent(uri, sid);
				Files.copy(new InputSupplier<InputStream>() {

					@Override
					public InputStream getInput() throws IOException {
						return tei;
					}
				}, corpus);
			}

			// Now, run the EPUB stylesheet
			// TODO cache the saxon stuff
			final XsltTransformer transformer;
			if (xsluri == null) {
				transformer = getTeiToEpub().load();
			} else {
				logger.info("Loading stylesheet from " + xsluri);
				final StreamSource xsltSource = new StreamSource(xsluri.toURL().openStream());
				final XsltExecutable xsltExecutable = xsltProcessor.newXsltCompiler().compile(xsltSource);
				transformer = xsltExecutable.load();
			}
			transformer.setDestination(xsltProcessor.newSerializer(new File(
					workingDir, "output.xml")));
			transformer.setSource(new StreamSource(corpus));
			transformer.transform();

			return RESTUtils
					.attachmentResponse(
							DefaultFilenamePolicy.INSTANCE.translate(rootObject.getGeneric().getProvided().getTitle().get(0))
							+ ".epub").type("application/epub+zip")
					.entity(new EPUBSerializer(workingDir, repository, corpus,
							sid)).build();
		} catch (final FileNotFoundException e) {
			throw new WebApplicationException(e);
		} catch (final IOException e) {
			throw new WebApplicationException(e);
		} catch (final SaxonApiException e) {
			throw new WebApplicationException(e);
		}
	}

}
