package info.textgrid.services.aggregator.epub;

import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.AuthFault;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.IoFault;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.MetadataParseFault;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.ObjectNotFoundFault;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.ProtocolNotImplementedFault;
import info.textgrid.services.aggregator.ITextGridRep;
import info.textgrid.utils.linkrewriter.ConfigurableXMLRewriter;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.Deque;
import java.util.LinkedList;
import java.util.Map.Entry;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.StreamingOutput;
import javax.xml.stream.XMLStreamException;

import org.apache.cxf.helpers.FileUtils;
import org.xml.sax.SAXException;

import com.google.common.io.ByteStreams;
import com.google.common.io.Files;

public class EPUBSerializer implements StreamingOutput {

	private final File workingDir;
	private final ITextGridRep repository;
	private final File corpus;
	private final String sid;



	public EPUBSerializer(final File workingDir, final ITextGridRep repository,
			final File corpus, final String sid) {
		super();
		this.workingDir = workingDir;
		this.repository = repository;
		this.corpus = corpus;
		this.sid = sid;
	}



	@Override
	public void write(final OutputStream output) throws IOException,
	WebApplicationException {
		final ZipOutputStream zip = new ZipOutputStream(output);
		try {

			// first entry is the uncompressed mimetype marker
			new MimeTypeEntry("application/epub+zip").writeTo(zip);

			// now filtered filesystem structure. FIXME add rewriting +
			// images
			final File mimeTypeFile = new File(workingDir, "mimetype");
			final URI base = workingDir.toURI();
			final File opsDir = new File(workingDir, "OPS");
			final URI ops = opsDir.toURI();
			final OPFManifest manifest = new OPFManifest(new File(opsDir,
					"content.opf"));
			final ConfigurableXMLRewriter xhtmlRewriter = new ConfigurableXMLRewriter(
					manifest.getImportMapping(), true);
			xhtmlRewriter.configure(URI.create("internal:html#html"));
			final ConfigurableXMLRewriter opfRewriter = new ConfigurableXMLRewriter(
					manifest.getImportMapping(), true);
			opfRewriter.configure(URI.create("internal:epub#opf"));

			final Deque<File> queue = new LinkedList<File>();
			queue.push(workingDir);
			while (!queue.isEmpty()) {
				final File directory = queue.pop();

				for (final File child : directory.listFiles()) {
					// filter stuff we don't want:
					if (child.equals(mimeTypeFile) || child.equals(corpus)) {
						continue;
					}

					String name = base.relativize(child.toURI()).getPath();
					if (child.isDirectory()) {
						queue.push(child);
						if (!name.endsWith("/")) {
							name = name.concat("/");
						}
						zip.putNextEntry(new ZipEntry(name));
					} else {
						zip.putNextEntry(new ZipEntry(name));

						if (Files.getFileExtension(name).equals("html")) {
							xhtmlRewriter.rewrite(
									new FileInputStream(child), zip);
						} else if (Files.getFileExtension(name).equals("opf")) {
							opfRewriter.rewrite(new FileInputStream(child),
									zip);
						} else {
							Files.copy(child, zip);
						}
						zip.closeEntry();
					}
				}
			}

			// now we need to add those files that are referenced by
			// absolute URI in the manifest
			for (final Entry<URI, String> externalItem : manifest.externalItems.entrySet()) {
				final String pseudoFileName = base.relativize(new File(opsDir, manifest.getFileName(externalItem.getKey().toString())).toURI()).getPath();
				zip.putNextEntry(new ZipEntry(pseudoFileName));
				ByteStreams.copy(
						repository.getContent(externalItem.getKey(), sid), zip);
				zip.closeEntry();
			}
		} catch (final SAXException e) {
			throw new WebApplicationException(e);
		} catch (final URISyntaxException e) {
			throw new WebApplicationException(e);
		} catch (final XMLStreamException e) {
			throw new WebApplicationException(e);
		} catch (final ObjectNotFoundFault e) {
			throw new WebApplicationException(e);
		} catch (final MetadataParseFault e) {
			throw new WebApplicationException(e);
		} catch (final IoFault e) {
			throw new WebApplicationException(e);
		} catch (final ProtocolNotImplementedFault e) {
			throw new WebApplicationException(e);
		} catch (final AuthFault e) {
			throw new WebApplicationException(e);
		} finally {
			zip.close();
			FileUtils.removeDir(workingDir);
		}

	}

}
