package info.textgrid.services.aggregator.html;

import info.textgrid.namespaces.metadata.core._2010.ObjectType;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.AuthFault;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.IoFault;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.MetadataParseFault;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.ObjectNotFoundFault;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.ProtocolNotImplementedFault;
import info.textgrid.services.aggregator.ITextGridRep;
import info.textgrid.services.aggregator.ITextGridRep.TGOSupplier;
import info.textgrid.services.aggregator.TextGridRepProvider;

import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.text.MessageFormat;
import java.util.concurrent.ExecutionException;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.servlet.ServletContext;
import javax.ws.rs.GET;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.Request;
import javax.ws.rs.core.Response;
import javax.xml.transform.stream.StreamSource;

import net.sf.saxon.s9api.Processor;
import net.sf.saxon.s9api.SaxonApiException;
import net.sf.saxon.s9api.XsltCompiler;
import net.sf.saxon.s9api.XsltExecutable;

import org.apache.cxf.jaxrs.model.wadl.Description;

import com.google.common.base.Optional;
import com.google.common.cache.Cache;
import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.RemovalListener;
import com.google.common.cache.RemovalNotification;

@Path("/html")
@Description("Creates an HTML representation of the given TEI document, or aggregation of TEI documents. This is currently extremely experimental and probably broken.")
public class HTML {


	private static final String TO_HTML_XSL = "/WEB-INF/stylesheets/db2xhtml.xsl";

	ITextGridRep repository = TextGridRepProvider.getInstance();

	final Logger logger = Logger
			.getLogger("info.textgrid.services.aggregator.html.HTML");

	private XsltExecutable toHtml;
	final Processor xsltProcessor;

	@Context
	private ServletContext servlet;

	private Cache<URI, XsltExecutable> stylesheets;

	XsltExecutable getToHtml() {
		if (toHtml == null) {
			try {
				final URL stylesheet = servlet.getResource(TO_HTML_XSL);
				final XsltCompiler xsltCompiler = xsltProcessor
						.newXsltCompiler();
				toHtml = xsltCompiler.compile(new StreamSource(stylesheet
						.toString()));
			} catch (final MalformedURLException e) {
				throw new IllegalStateException(e);
			} catch (final SaxonApiException e) {
				throw new IllegalStateException(e);
			}
		}
		return toHtml;
	}

	public HTML(final ITextGridRep repository) throws IOException {
		this.repository = repository;
		xsltProcessor = new Processor(false);
		xsltProcessor.getUnderlyingConfiguration().setURIResolver(
				new TGUriResolver(repository));
		// xsltProcessor.getUnderlyingConfiguration().setAllowExternalFunctions(
		// false); // we run external stylesheets
		xsltProcessor.getUnderlyingConfiguration().setCompileWithTracing(true);
		stylesheets = CacheBuilder.newBuilder().maximumSize(50).softValues()
				.removalListener(new RemovalListener<URI, XsltExecutable>() {

					@Override
					public void onRemoval(
							final RemovalNotification<URI, XsltExecutable> notification) {
						logger.info(MessageFormat
								.format("Removed stylesheet {0} from cache, reason: {1}",
										notification.getKey(),
										notification.getCause()));
					}
				})
.build(new CacheLoader<URI, XsltExecutable>() {

					@Override
					public XsltExecutable load(final URI url) throws Exception {
						final XsltCompiler compiler = xsltProcessor
								.newXsltCompiler();
						try {
							final XsltExecutable executable = compiler
									.compile(new StreamSource(url.toString()));
							logger.log(Level.INFO,
									"Successfully loaded stylesheet {0}", url);
							return executable;
						} catch (final Exception e) {
							logger.log(Level.SEVERE, MessageFormat.format(
									"Failed to load stylesheet {0}", url), e);
							throw e;
						}
					}
				});

	}

	/**
	 * Returns an appropriate stylesheet for the given URI.
	 *
	 * Basically, we try the following options in order:
	 * <ol>
	 * <li>The stylesheet is cached -> return the cached version.
	 * <li>The stylesheet is public or external -> load & cache it.
	 * <li>The stylesheet is non-public TextGrid internal -> load & do not cache
	 * it.
	 * </ol>
	 *
	 * @param uri
	 *            the URI of the stylesheet to load
	 * @param sid
	 *            the session ID to use, if present
	 * @param forceLoad
	 *            do not use a cached version even if present.
	 * @throws IOException
	 *             if an error occurs reading the stylesheet.
	 * @throws SaxonApiException
	 *             if saxon fails to compile the stylesheet.
	 */
	protected XsltExecutable getStylesheet(final URI uri,
			final Optional<String> sid,
			final boolean forceLoad) throws SaxonApiException, IOException {
		XsltExecutable executable = null;

		// (1) try cached version, if it exists
		if (!forceLoad) {
			executable = stylesheets.getIfPresent(uri);
		}

		if (executable == null) {

			final XsltCompiler compiler = xsltProcessor.newXsltCompiler();
			if (TGUriResolver.isResolveable(uri)) {

				// (2/3) it's a TextGrid object, load it from TG-crud.
				final TGOSupplier<InputStream> xsltSupplier = repository.read(
						uri, sid.orNull());
				executable = compiler.compile(new StreamSource(xsltSupplier
						.getInput(), uri.toString()));

				if (isPublic(xsltSupplier.getMetadata())) {
					// (2) it's public -> we can cache it.
					stylesheets.put(uri, executable);
					logger.log(Level.INFO, "Cached public stylesheet {0}", uri);
				} else {
					logger.log(Level.INFO, "Loaded private stylesheet {0}", uri);
				}
			} else {
				// (2) it's non-TextGrid -- load & cache it.
				executable = compiler.compile(new StreamSource(uri.toString()));
				stylesheets.put(uri, executable);
				logger.log(Level.INFO, "Cached external stylesheet {0}", uri);
			}
		} else {
			logger.log(Level.INFO, "Reusing cached stylesheed {0}", uri);
		}

		return executable;
	}

	private static boolean isPublic(final ObjectType metadata) {
		try {
			return metadata.getGeneric().getGenerated().getAvailability()
					.contains("public");
		} catch (final NullPointerException e) {
			return false;
		}
	}

	@GET
	@Path(value = "/{object}")
	@Produces(value = "text/xml")
	public Response get(
			@Description("The TextGrid URI of the TEI document or aggregation to transform")
			@PathParam("object") final URI uri,
			@Description("If given, an alternative XSLT stylesheet to use")
			@QueryParam("stylesheet") final URI xsluri,
			@Description("If true, check for an <?xsl-stylesheet?> processing instruction in the document to render")
			@QueryParam("pi") final boolean pi,
			@Description("If true and a stylesheet has been given, force reloading the stylesheet, do not cache") @QueryParam("refreshStylesheet") final boolean refreshStylesheet,
			@Description("Session ID to access protected resources")
			@QueryParam("sid") final String sid,
			@Description("If true, pass the information the stylesheet that its result will be embedded into some website") @QueryParam("embedded") final boolean embedded,
			@Description("URL of the CSS that should be referenced in the HTML that is created")
			@QueryParam("css") final URI css,
			@Description("The requested content type. E.g., text/html or text/xml") @QueryParam("mediatype") final String mediaType,
			@Context final Request request) throws ObjectNotFoundFault,
			MetadataParseFault, IoFault, AuthFault,
			ProtocolNotImplementedFault, WebApplicationException, IOException,
			SaxonApiException, ExecutionException {

		logger.fine("HTML called for root object: " + uri);

		final HTMLWriter writer = new HTMLWriter(this, uri, xsluri,
				refreshStylesheet, pi, embedded, css, sid, mediaType, request);
		return writer.createResponse().build();
	}

}
