package info.textgrid.services.aggregator.html;

import info.textgrid.namespaces.metadata.core._2010.ObjectType;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.AuthFault;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.IoFault;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.MetadataParseFault;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.ObjectNotFoundFault;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.ProtocolNotImplementedFault;
import info.textgrid.services.aggregator.GenericExceptionMapper;
import info.textgrid.services.aggregator.ITextGridRep;
import info.textgrid.services.aggregator.ITextGridRep.TGOSupplier;
import info.textgrid.services.aggregator.teicorpus.TEICorpusSerializer;
import info.textgrid.utils.export.filenames.DefaultFilenamePolicy;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.text.MessageFormat;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.CacheControl;
import javax.ws.rs.core.Request;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.Response.ResponseBuilder;
import javax.ws.rs.core.Response.Status;
import javax.ws.rs.core.StreamingOutput;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Source;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.stream.StreamSource;

import net.sf.saxon.s9api.QName;
import net.sf.saxon.s9api.SaxonApiException;
import net.sf.saxon.s9api.Serializer;
import net.sf.saxon.s9api.XdmAtomicValue;
import net.sf.saxon.s9api.XsltExecutable;
import net.sf.saxon.s9api.XsltTransformer;

import com.google.common.base.Optional;
import com.google.common.base.Stopwatch;
import com.google.common.base.Throwables;
import com.google.common.io.ByteStreams;
import com.google.common.io.FileBackedOutputStream;

/**
 * The essential steps:
 *
 * <ol>
 * <li>Construct the transformation source.
 * <p>
 * Depends on different factors: If it's an aggregation, we need to recursively
 * construct a teiCorpus first. If we need to parse the
 * {@code <?xsl-stylesheet?>} processing instruction, we need to at least
 * partially parse the document twice, so we download it to a temporary buffer.
 * Otherwise, the source could be the result stream of the call to TG-crud.
 * </p>
 * <p>
 * Includes some, potentially longish, requests to TextGridRep or other sources
 * that may cause exceptions.
 * </p>
 * </li>
 * <li>Load and configure the stylesheet.
 * <p>
 * Typically, the order is argument-specified > associated (via pi) > default
 * stylesheet, with the associated stylesheet only checked on request.
 * </p>
 * </li>
 * <li>Perform the transformation and output the result.
 * <p>
 * This step essentially requires the output stream to be present.
 * </p>
 * </li>
 * </ol>
 *
 * @author Thorsten Vitt <thorsten.vitt@uni-wuerzburg.de>
 *
 */
public class HTMLWriter implements StreamingOutput {

	private final HTML service;

	private final static Logger logger = Logger.getLogger(HTMLWriter.class.getCanonicalName());

	// Options
	private final URI rootURI;
	private Optional<URI> stylesheetURI = Optional.absent();
	private boolean refreshStylesheet = false;
	private boolean readStylesheetPI = false;
	private boolean embedded = false;

	private Optional<URI> css = Optional.absent();
	private Optional<String> sid;

	// detected and extracted
	private enum SourceType {
		UNKNOWN, XML, AGGREGATION
	}

	private SourceType sourceType = SourceType.UNKNOWN;

	private final ITextGridRep repository;

	private ObjectType metadata;

	private Optional<URI> associatedStylesheet = Optional.absent();

	private Source source = null;

	private XsltTransformer transformer;

	private final Stopwatch stopwatch;

	private String actualStylesheet;

	private TGOSupplier<InputStream> content;

	private Request request;

	private Optional<String> requestedMediaType;

	// Constructor and configuration

	public HTMLWriter(final HTML service, final URI rootURI) {
		stopwatch = new Stopwatch();
		stopwatch.start();
		logger.log(Level.INFO, "Starting HTML export for {0}", rootURI);
		this.service = service;
		this.rootURI = rootURI;
		this.repository = service.repository;
	}

	public HTMLWriter(final HTML service, final URI rootURI,
			final URI stylesheetURI, final boolean refreshStylesheet,
			final boolean readStylesheetPI, final boolean embedded,
			final URI css, final String sid, final String mediaType,
			final Request request) {

		this(service, rootURI);

		this.stylesheetURI = Optional.fromNullable(stylesheetURI);
		this.refreshStylesheet = refreshStylesheet;
		this.readStylesheetPI = readStylesheetPI;
		this.embedded = embedded;
		this.css = Optional.fromNullable(css);
		this.sid(sid);
		this.requestedMediaType = Optional.fromNullable(mediaType);
		this.request = request;
	}

	public HTMLWriter sid(final String sid) {
		if (sid == null || sid.isEmpty()) {
			this.sid = Optional.absent();
		} else {
			this.sid = Optional.of(sid);
		}
		return this;
	}

	public HTMLWriter stylesheet(final URI uri) {
		this.stylesheetURI = Optional.fromNullable(uri);
		return this;
	}

	public HTMLWriter refresh(final boolean refresh) {
		this.refreshStylesheet = refresh;
		return this;
	}

	public HTMLWriter embedded(final boolean embedded) {
		this.readStylesheetPI = embedded;
		if (embedded) {
			this.sourceType = SourceType.XML;
		}
		return this;
	}

	public HTMLWriter css(final URI css) {
		this.css = Optional.fromNullable(css);
		return this;
	}

	protected HTMLWriter loadSource() throws ObjectNotFoundFault,
			MetadataParseFault,
			IoFault, ProtocolNotImplementedFault, AuthFault, IOException {

		metadata = getContent().getMetadata();
		final String format = metadata.getGeneric().getProvided().getFormat();
		if (format.contains("aggregation")) {
			sourceType = SourceType.AGGREGATION;
		} else if (format.matches("^text/.*xml.*$")) {
			sourceType = SourceType.XML;
		} else {
			final String errorMsg = MessageFormat
					.format("The HTML export can only convert aggregations or XML documents to EPUB, however, the document {0} you referred to has the MIME type {1}.",
							rootURI, format);
			throw new WebApplicationException(
					GenericExceptionMapper.toResponse(
							Status.UNSUPPORTED_MEDIA_TYPE, errorMsg, ""));
		}

		if (sourceType == SourceType.AGGREGATION) {
			final TEICorpusSerializer corpusSerializer = new TEICorpusSerializer(
					metadata, false, sid.orNull());
			final FileBackedOutputStream corpusBuffer = new FileBackedOutputStream(
					1024 * 1024, true);
			corpusSerializer.write(corpusBuffer);
			corpusBuffer.close();
			this.source = new StreamSource(corpusBuffer.getSupplier()
					.getInput());
		} else if (sourceType == SourceType.XML && readStylesheetPI) {
			final FileBackedOutputStream xmlBuffer = new FileBackedOutputStream(
					1024 * 1024, true);
			ByteStreams.copy(getContent(), xmlBuffer);
			detectEmbeddedStylesheet(xmlBuffer.getSupplier().getInput());
			this.source = new StreamSource(xmlBuffer.getSupplier().getInput());
		} else {
			this.source = new StreamSource(getContent().getInput(),
					rootURI.toString());
		}
		logger.log(Level.INFO, MessageFormat.format("Fetched source for {0}, type={1}, after {2}", rootURI, sourceType, stopwatch.toString()));

		return this;
	}

	private TGOSupplier<InputStream> getContent() {
		if (content == null) {
			content = service.repository.read(rootURI, sid.orNull());
			logger.info(MessageFormat.format("Fetched source for {0} up to metadata after {1}", rootURI, stopwatch));
		}
		return content;
	}

	private void detectEmbeddedStylesheet(final InputStream input) {
		try {
			final Source associatedStylesheet = TransformerFactory
					.newInstance().getAssociatedStylesheet(
							new StreamSource(input, rootURI.toString()), null,
							null, null);
			this.associatedStylesheet = Optional.of(new URI(
					associatedStylesheet.getSystemId()));

		} catch (final TransformerConfigurationException e) {
			logger.log(Level.WARNING, "Failed to load stylesheet from <?xsl-stylesheet?> declaration", e);
		} catch (final TransformerFactoryConfigurationError e) {
			logger.log(Level.WARNING, "Failed to load stylesheet from <?xsl-stylesheet?> declaration", e);
		} catch (final URISyntaxException e) {
			logger.log(Level.WARNING, "Failed to load stylesheet from <?xsl-stylesheet?> declaration", e);
		}
	}

	private XsltExecutable getStylesheet() throws SaxonApiException,
			IOException {
		if (stylesheetURI.isPresent()) {
			actualStylesheet = stylesheetURI.get().toString() + " (explicit)";
			return service.getStylesheet(stylesheetURI.get(), sid,
					refreshStylesheet);
		} else if (associatedStylesheet.isPresent()) {
			actualStylesheet = associatedStylesheet.get().toString() + " (associated)";
			return service.getStylesheet(associatedStylesheet.get(), sid,
					refreshStylesheet);
		} else {
			actualStylesheet = "(internal)";
			return service.getToHtml();
		}
	}

	protected HTMLWriter loadStylesheet() throws SaxonApiException,
			IOException, ObjectNotFoundFault, MetadataParseFault, IoFault,
			ProtocolNotImplementedFault, AuthFault {
		if (source == null) {
			loadSource();
		}

		final XsltTransformer transformer = getStylesheet().load();
		if (sid.isPresent()) {
			transformer.setURIResolver(new TGUriResolver(repository, sid));
		}
		transformer.setParameter(new QName("graphicsURLPattern"),
				new XdmAtomicValue(repository.getCRUDRestEndpoint()
						+ "/@URI@/data"
						+ ((sid == null || "".equals(sid)) ? ""
								: ("?sessionId=" + sid))));
		if (css.isPresent()) {
			transformer.setParameter(new QName("cssFile"), new XdmAtomicValue(
					css.get()));
		}
		transformer.setParameter(new QName("embedded"), new XdmAtomicValue(
				embedded));

		transformer.setSource(source);

		if (requestedMediaType.isPresent()) {
			transformer.getUnderlyingController().setOutputProperty(
					OutputKeys.MEDIA_TYPE, requestedMediaType.get());
		}

		this.transformer = transformer;
		logger.log(Level.INFO, MessageFormat.format("Prepared XSLT stylesheet {1} for {0} after {2}", rootURI, actualStylesheet, stopwatch.toString()));
		return this;
	}

	@Override
	public void write(final OutputStream out) throws IOException,
			WebApplicationException {
		try {
			if (source == null) {
				loadSource();
			}
			if (transformer == null) {
				loadStylesheet();
			}
			logger.log(Level.INFO, MessageFormat.format("Ready for transformation of {0} after {1}", rootURI, stopwatch.toString()));
			final Serializer serializer = service.xsltProcessor
					.newSerializer(out);
			transformer
.setDestination(serializer);
			transformer.transform();

		} catch (final Exception e) {
			logger.log(Level.SEVERE, MessageFormat.format("Transformation of {0} failed ({2}) after {1}", rootURI, stopwatch.toString(), e.getMessage()));
			Throwables.propagateIfPossible(e, IOException.class,
					WebApplicationException.class);
			Throwables.propagate(e);
		}
		logger.log(Level.INFO, MessageFormat.format("Finished and delivered transformation of {0} after {1}", rootURI, stopwatch.toString()));
		stopwatch.stop();
	}

	public ResponseBuilder createResponse() throws ObjectNotFoundFault, MetadataParseFault, IoFault, ProtocolNotImplementedFault, AuthFault, IOException, SaxonApiException {
		if (!refreshStylesheet && request != null) {
			final ResponseBuilder builder = request.evaluatePreconditions(getContent().getMetadata().getGeneric().getGenerated().getLastModified().toGregorianCalendar().getTime());
			if (builder != null) {
				logger.info(MessageFormat.format("Aborting: client already has HTML for {0} (after {1})", rootURI, stopwatch));
				return builder;
			} else {
				logger.info("Creating new transformation.");
			}
		}

		loadSource();
		loadStylesheet();

		final ResponseBuilder builder = Response.ok();
		if (requestedMediaType.isPresent()) {
			builder.type(requestedMediaType.get());
		} else {
			final Properties outputProperties = transformer
					.getUnderlyingController().getOutputProperties();
			builder.type(outputProperties.getProperty(OutputKeys.MEDIA_TYPE,
					"text/html"));
		}

		builder.lastModified(getContent().getMetadata().getGeneric().getGenerated().getLastModified().toGregorianCalendar().getTime());
		final CacheControl cacheControl = new CacheControl();
		cacheControl.setPrivate(sid.isPresent());
		cacheControl.setMaxAge(86400);	// one day
		builder.cacheControl(cacheControl);
		builder.header("Content-Disposition", "inline;filename=\"" + DefaultFilenamePolicy.INSTANCE.getFilename(getContent().getMetadata(), false) + ".html\"");
		builder.entity(this);
		return builder;
	}

}
