package info.textgrid.services.aggregator.teicorpus;

import info.textgrid.namespaces.metadata.core._2010.MetadataContainerType;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.AuthFault;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.IoFault;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.MetadataParseFault;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.ObjectNotFoundFault;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.TGCrudService;
import info.textgrid.services.aggregator.AggregationTreeWalker;
import info.textgrid.services.aggregator.ITextGridRep;
import info.textgrid.services.aggregator.RESTUtils;

import java.net.URI;
import java.net.URISyntaxException;
import java.text.MessageFormat;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.ws.rs.DefaultValue;
import javax.ws.rs.GET;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.Response.ResponseBuilder;
import javax.ws.rs.core.Response.Status;

import org.apache.cxf.jaxrs.model.wadl.Description;
import org.apache.cxf.jaxrs.model.wadl.Descriptions;
import org.apache.cxf.jaxrs.model.wadl.DocTarget;

/*
 * Generating a TEIcorpus document from a (single) aggregation
 * 
 * Header:
 * 
 * Header could be generated in the same way as in the metadata editor, i.e.
 * paste together the ancestor metadata records (including matching work
 * records) and then run an XSLT stylesheet.
 * 
 * Children's Header:
 * 
 * For descendant aggregations, we still need a header. Either we consider this
 * redundant, or we preserve the metadata records we copied together for the
 * respective ancestor levels, append the current level & run the
 * transformation.
 * 
 * Body:
 * 
 * Naïve approach: Load and recursively process the aggregations, check each
 * one's content type Probably better: SearchClient has a listAggregation()
 * method that is also used by the navigator and that returns metadata records.
 * We could use this, so it's only one large request per aggregation. Should be
 * reasonably fast.
 */
@Description("Creates a TEI corpus of all the TEI documents (recursively) aggregated by the given aggregation")
@Path("/teicorpus")
public class TEICorpus {

	private final ITextGridRep repository;

	final Logger logger = Logger.getLogger("info.textgrid.services.aggregator");

	public TEICorpus(final ITextGridRep repository) {
		this.repository = repository;
	}


	@GET
	@Path(value="/{aggregation}")
	@Produces("text/xml")
	@Descriptions({
			@Description(target=DocTarget.METHOD, value="Creates a TEI corpus of all the TEI documents (recursively) aggregated by the given aggregation"),
			@Description(target=DocTarget.RETURN, value="TEI corpus document")
	})
	public Response get(@Description("TextGrid URI of the root aggregation") @PathParam("aggregation") final URI uri,
			@Description("Whether to generate a Content-Disposition: attachment header") @QueryParam("attach") @DefaultValue("true") final boolean attach,
			@Description("If true, no intermediate TEI corpus documents will be generated for intermediate aggregations, hierarchical structure will be lost") @QueryParam("flat") @DefaultValue("false") final boolean flat,
			@Description("Session id for accessing restricted resources") @QueryParam("sid") final String sid)
					throws URISyntaxException, ObjectNotFoundFault, MetadataParseFault, IoFault, AuthFault {
		logger.fine("TEIcorpus called for root aggregation: " + uri);
		final TGCrudService crud = repository.getCRUDService();
		logger.finest("Yo, clients are there.");
		final MetadataContainerType rootAggregationMetadata = crud
				.readMetadata(sid, null, uri.toString());
		logger.finer("CRUD request for root aggregation successful");
		final AggregationTreeWalker serializer = new TEICorpusSerializer(
				rootAggregationMetadata.getObject(), flat, sid);
		final String format = rootAggregationMetadata.getObject().getGeneric().getProvided().getFormat();
		if (!format.contains("aggregation")) {
			logger.log(Level.SEVERE, "The requested object {0} is a {1}, not an aggregation, and thus cannot be aggregated.", new Object[] {uri,format});
			throw new WebApplicationException(javax.ws.rs.core.Response
					.status(Status.BAD_REQUEST)
					.entity(MessageFormat.format(
							"The requested object {0} is a {1}, not an aggregation, and thus cannot be aggregated.", uri,
							format)).build());
		}
		final ResponseBuilder builder = Response.ok(serializer, "text/xml");
		if (attach) {
			final String fileName = rootAggregationMetadata.getObject()
					.getGeneric().getProvided().getTitle().get(0)
					+ (flat ? ".flat" : "") + ".xml";
			RESTUtils.addAttachmentFilename(builder, fileName);
		}
		return builder.build();

		// final Response response = search.listAggregation(uri.toString());
		// for (final ResultType result : response.getResult()) {
		// final String contentType =
		// result.getObject().getGeneric().getProvided().getFormat();
		// if (contentType.contains("aggregation")) {
		// // TODO deal with aggregation
		// } else if ("text/xml".equals(contentType)) {
		// // TODO deal with
		// }
		// }
		// return null;
	}

}
