package info.textgrid.services.aggregator.teicorpus;

import info.textgrid.namespaces.metadata.agent._2010.AgentType;
import info.textgrid.namespaces.metadata.core._2010.MetadataContainerType;
import info.textgrid.namespaces.metadata.core._2010.ObjectType;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.AuthFault;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.IoFault;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.MetadataParseFault;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.ObjectNotFoundFault;
import info.textgrid.namespaces.middleware.tgcrud.services.tgcrudservice.ProtocolNotImplementedFault;
import info.textgrid.services.aggregator.AggregationTreeWalker;
import info.textgrid.services.aggregator.TextGridRepProvider;

import java.io.IOException;
import java.io.OutputStream;
import java.text.MessageFormat;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.activation.DataHandler;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.StreamingOutput;
import javax.xml.stream.EventFilter;
import javax.xml.stream.XMLEventFactory;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLEventWriter;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.XMLEvent;
import javax.xml.transform.stax.StAXResult;
import javax.xml.ws.Holder;

import com.google.common.base.Joiner;
import com.google.common.base.Throwables;
import com.google.common.collect.Sets;

public class TEICorpusSerializer extends AggregationTreeWalker implements
		StreamingOutput {

	private static final String TEI_CORPUS = "teiCorpus";
	private static final String TEI_NS = "http://www.tei-c.org/ns/1.0";
	private final Logger logger = Logger.getLogger("info.textgrid.services.aggregator.teicorpus.serializer");
	

	private XMLOutputFactory outputFactory;
	private XMLEventFactory eventFactory;
	private XMLEventWriter writer;
	private final ObjectType rootObject;
	private XMLInputFactory inputFactory;
	private final Set<String> seen = Sets.newHashSet();
	private static final EventFilter NO_DOCUMENT_NODE = new EventFilter() {

		@Override
		public boolean accept(final XMLEvent event) {
			return !(event.isStartDocument() || event.isEndDocument());
		}
	};
	private TEIHeaderStack stack;
	private boolean handled;
	private boolean flat;
	private int level = 0;
	private static String toString(final ObjectType object) {
		return Joiner.on(" / ").join(object.getGeneric().getProvided().getTitle()) + " (" + object.getGeneric().getGenerated().getTextgridUri().getValue() + ", " +
			object.getGeneric().getProvided().getFormat() + ")";
	}

	public TEICorpusSerializer(final ObjectType object)
			throws ObjectNotFoundFault, MetadataParseFault, IoFault, AuthFault {
		this.rootObject = object;
		setRepository(TextGridRepProvider.getInstance());
		logger.info("Starting TEIcorpus serialization for " + toString(object));
	}

	public TEICorpusSerializer(final ObjectType object, final boolean flat,
			final String sid)
			throws ObjectNotFoundFault, MetadataParseFault, IoFault, AuthFault {
		this(object);
		this.flat = flat;
		this.setSid(sid);
	}

	@Override
	public void write(final OutputStream output) throws IOException, WebApplicationException {
		outputFactory = XMLOutputFactory.newInstance();
		eventFactory = XMLEventFactory.newInstance();
		outputFactory.setProperty(XMLOutputFactory.IS_REPAIRING_NAMESPACES, true);
		inputFactory = XMLInputFactory.newInstance();
		try {
			writer = outputFactory.createXMLEventWriter(output);
			writer.setDefaultNamespace(TEI_NS);
			writer.add(eventFactory.createStartDocument());

			walk(rootObject, false);

			writer.close();
		} catch (final XMLStreamException e) {
			logger.log(Level.SEVERE, "XML Error during serialization", e);
			throw new WebApplicationException(e);
		}
	}

	@Override
	protected boolean walk(final ObjectType object, final boolean again) {
		if (seen.contains(object.getGeneric().getGenerated().getTextgridUri().getValue()))
			writeSkip(object);

		handled = super.walk(object, again);

		if (!handled)
			if (object.getGeneric().getProvided().getFormat()
						.equals("text/xml"))
			walkXML(object);
		else
			writeSkip(object);
		return true;
	}

	@Override
	protected void walkAggregation(final ObjectType object, final boolean again) {
		logger.fine("[Processing aggregation " + toString(object));
		level++;
		if (stack == null)
			try {
				stack = new TEIHeaderStack(object, getSearchClient(isPublic(object)));
			} catch (Exception e) {
				throw Throwables.propagate(e);
			}
		else
			stack.push(object);

		if (!flat || level <= 1) {
		try {
			writer.add(eventFactory.createStartElement("", TEI_NS, TEI_CORPUS));
			writer.add(eventFactory.createAttribute("corresp", object
					.getGeneric().getGenerated().getTextgridUri().getValue()));
		} catch (XMLStreamException e) {
			throw new IllegalStateException(e);
		}
		
		// writeHeader(object);
		stack.writeHeader(new StAXResult(new XMLEventFilter(writer) {

			@Override
			public void add(final XMLEvent event) throws XMLStreamException {
				if (!(event.isStartDocument() || event.isEndDocument()))
					super.add(event);
			}

		}
		));
		}

		super.walkAggregation(object, again);

		if (!flat || level <= 1)
			try {
				writer.add(eventFactory.createEndElement("", TEI_NS, TEI_CORPUS));
			} catch (XMLStreamException e) {
				throw new IllegalStateException(e);
			}

		level--;
		logger.fine("Finished aggregation " + toString(object) + "]");
	}

	@SuppressWarnings("unused")
	private void writeHeader(final ObjectType object) throws XMLStreamException {
		writer.add(eventFactory.createStartElement("",  TEI_NS, "teiHeader"));
		writer.add(eventFactory.createComment("This is a (temporary) header generated from aggregation metadata."));
		writer.add(eventFactory.createStartElement("",  TEI_NS, "fileDesc"));
		writer.add(eventFactory.createStartElement("",  TEI_NS, "titleStmt"));
		for (final String title : object.getGeneric().getProvided().getTitle())
			writeSimpleTEIElement("title", title);
		if (object.getEdition() != null)
			for (AgentType agent : object.getEdition().getAgent()) {
				writer.add(eventFactory.createStartElement("",  TEI_NS, "respStmt"));
				writeSimpleTEIElement("resp", agent.getRole().toString());
				writer.add(eventFactory.createStartElement("",  TEI_NS, "author"));
				if (agent.getId() != null)
					writer.add(eventFactory.createAttribute("key", agent.getId()));
				writer.add(eventFactory.createCharacters(agent.getValue()));
				writer.add(eventFactory.createEndElement("",  TEI_NS, "author"));
				writer.add(eventFactory.createEndElement("",  TEI_NS, "respStmt"));
			}
		writer.add(eventFactory.createEndElement("",  TEI_NS, "titleStmt"));
		writer.add(eventFactory.createStartElement("",  TEI_NS, "publicationStmt"));
		writer.add(eventFactory.createStartElement("",  TEI_NS, "idno"));
		writer.add(eventFactory.createAttribute("type", "textgrid"));
		writer.add(eventFactory.createCharacters(object.getGeneric().getGenerated().getTextgridUri().getValue()));
		writer.add(eventFactory.createEndElement("",  TEI_NS, "idno"));
		writer.add(eventFactory.createEndElement("",  TEI_NS, "publicationStmt"));
		writer.add(eventFactory.createEndElement("",  TEI_NS, "fileDesc"));
		writer.add(eventFactory.createEndElement("",  TEI_NS, "teiHeader"));
	}

	private void writeSimpleTEIElement(final String element, final String content) throws XMLStreamException {
		writer.add(eventFactory.createStartElement("",  TEI_NS, element));
		writer.add(eventFactory.createCharacters(content));
		writer.add(eventFactory.createEndElement("",  TEI_NS, element));
	}

	protected void walkXML(final ObjectType object) {
		logger.fine("Processing XML " + toString(object));
		final Holder<MetadataContainerType> mdHolder = new Holder<MetadataContainerType>();
		final Holder<DataHandler> dhHolder = new Holder<DataHandler>();
		try {
			getRepository().getCRUDService().read(
					getSid(),
					"",
					object.getGeneric().getGenerated().getTextgridUri()
							.getValue(), mdHolder,
					dhHolder);
			final XMLEventReader reader = inputFactory.createFilteredReader(
					inputFactory.createXMLEventReader(dhHolder.value.getInputStream()),
					NO_DOCUMENT_NODE);

			// skip to document node:
			while (reader.hasNext() && !reader.peek().isStartElement())
				reader.next();

			writer.add(reader);
		} catch (final ObjectNotFoundFault e) {
			throw new WebApplicationException(e, Response.Status.NOT_FOUND);
		} catch (final MetadataParseFault e) {
			throw new WebApplicationException(e);
		} catch (final IoFault e) {
			throw new WebApplicationException(e);
		} catch (final ProtocolNotImplementedFault e) {
			throw new WebApplicationException(e);
		} catch (final AuthFault e) {
			throw new WebApplicationException(e, Response.Status.FORBIDDEN);
		} catch (final XMLStreamException e) {
			throw new WebApplicationException(e);
		} catch (final IOException e) {
			throw new WebApplicationException(e, Response.Status.SERVICE_UNAVAILABLE);
		}
	}

	/**
	 * Writes a comment for objects that are skipped.
	 */
	private void writeSkip(final ObjectType object) {
		logger.fine("Skipped " + toString(object));
		try {
			writer.add(eventFactory.createComment(MessageFormat.format(
					"Skipped {0}.", toString(object))));
		} catch (XMLStreamException e) {
			throw new IllegalStateException(e);
		}
	}

}
