package info.textgrid.services.aggregator.zip;

import info.textgrid._import.ImportObject;
import info.textgrid._import.RewriteMethod;
import info.textgrid.namespaces.metadata.core._2010.ObjectType;
import info.textgrid.services.aggregator.ITextGridRep;
import info.textgrid.services.aggregator.tree.AggregationTreeFactory;
import info.textgrid.utils.export.aggregations.Aggregation;
import info.textgrid.utils.export.aggregations.AggregationEntry;
import info.textgrid.utils.export.aggregations.IAggregation;
import info.textgrid.utils.export.aggregations.IAggregationEntry;
import info.textgrid.utils.export.filenames.DefaultFilenamePolicy;
import info.textgrid.utils.export.filenames.DefaultMetaFilenamePolicy;
import info.textgrid.utils.export.filenames.IFilenamePolicy;
import info.textgrid.utils.linkrewriter.ConfigurableXMLRewriter;
import info.textgrid.utils.linkrewriter.ImportMapping;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URI;
import java.text.MessageFormat;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.StreamingOutput;
import javax.xml.bind.JAXB;
import javax.xml.stream.XMLStreamException;

import com.google.common.base.Optional;
import com.google.common.base.Throwables;
import com.google.common.collect.ImmutableMap;
import com.google.common.io.ByteStreams;
import com.google.common.io.FileBackedOutputStream;

public class ZipResult implements StreamingOutput {

	private final ObjectType rootMetadata;
	private final ITextGridRep repository;
	private final String sid;
	private final IFilenamePolicy policy;
	private final IFilenamePolicy metaPolicy;
	private ImportMapping mapping;
	private final static Logger logger = Logger.getLogger(ZipResult.class.getCanonicalName());

	public ZipResult(final ObjectType rootObject, final ITextGridRep repository, final String sid) {
		this.rootMetadata = rootObject;
		this.repository = repository;
		this.sid = sid;
		this.policy = new DefaultFilenamePolicy();
		this.metaPolicy = new DefaultMetaFilenamePolicy(policy);
	}

	@Override
	public void write(final OutputStream output) throws IOException,
			WebApplicationException {
		logger.log(Level.INFO, "Starting ZIP export of {0}", rootMetadata
				.getGeneric().getGenerated().getTextgridUri().getValue());
		final long startTime = System.currentTimeMillis();
		final ZipOutputStream zip = new ZipOutputStream(output);
		zip.setComment(MessageFormat.format("# Exported from TextGrid -- www.textgrid.de\nRoot aggregation: {0}", rootMetadata.getGeneric().getGenerated().getTextgridUri().getValue()));

		final Aggregation root = AggregationTreeFactory.create(rootMetadata, repository, sid);

		logger.log(Level.INFO, "Built aggregation tree after {0,number} ms",
				System.currentTimeMillis() - startTime);

		mapping = new ImportMapping();
		addToMapping(mapping, root);
		writeAggregation(zip, root);

		// now serializing the mapping
		zip.putNextEntry(new ZipEntry(".INDEX.imex"));
		JAXB.marshal(mapping.toImportSpec(), zip);
		zip.closeEntry();
		zip.close();
		logger.log(Level.INFO, MessageFormat.format(
				"Finished exporting {0} after {1,number} ms", root,
				System.currentTimeMillis() - startTime));
	}

	// FIXME refactor -> Rewrite library
	private static ImmutableMap<String, String> REWRITE_CONFIGS = null;
	private static Optional<String> getRewriteConfig(final String contentType) {
		if (REWRITE_CONFIGS == null) {
			REWRITE_CONFIGS = ImmutableMap.<String, String>builder()
					.put("text/tg.aggregation+xml", "internal:textgrid#aggregation")
					.put("text/tg.edition+tg.aggregation+xml", "internal:textgrid#aggregation")
					.put("text/tg.collection+tg.aggregation+xml", "internal:textgrid#aggregation")
					.put("text/xsd+xml", "internal:schema#xsd")
					.put("text/linkeditorlinkedfile", "internal:tei#tei")
					.put("text/xml", "internal:tei#tei")
					.put("application/xhtml+xml", "internal:html#html")
					.build();
		}
		return Optional.fromNullable(REWRITE_CONFIGS.get(contentType));
	}

	/**
	 * Recursively adds the entry to the mapping.
	 * @param mapping
	 * @param childEntry
	 */
	private void addToMapping(final ImportMapping mapping, final AggregationEntry childEntry) {
		final Optional<String> rewriteConfig = getRewriteConfig(childEntry.getFormat());
		final ImportObject importObject = new ImportObject();
		importObject.setTextgridUri(childEntry.getTextGridURI().toString());
		importObject.setLocalData(policy.getFilename(childEntry, false).toString());
		importObject.setLocalMetadata(metaPolicy.getFilename(childEntry, false).toString());
		if (rewriteConfig.isPresent()) {
			importObject.setRewriteMethod(RewriteMethod.XML);
			importObject.setRewriteConfig(rewriteConfig.get());
		} else {
			importObject.setRewriteMethod(RewriteMethod.NONE);
		}
		mapping.add(importObject);

		if (childEntry instanceof IAggregation) {
			final IAggregation aggregation = (IAggregation) childEntry;
			for (final IAggregationEntry child : aggregation.getChildren()) {
				addToMapping(mapping, (AggregationEntry) child);
			}
		}
	}

	private void writeAggregation(final ZipOutputStream zip, final Aggregation aggregation) throws IOException {
		writeFile(zip, aggregation);
		final ZipEntry zipEntry = new ZipEntry(policy.getFilename(aggregation, true).toString());
		zip.putNextEntry(zipEntry);
		zipEntry.setTime(aggregation.getMetadata().getGeneric().getGenerated().getLastModified().toGregorianCalendar().getTimeInMillis());
		zip.closeEntry();

		for (final IAggregationEntry child : aggregation.getChildren()) {
			if (child instanceof Aggregation) {
				writeAggregation(zip, (Aggregation) child);
			} else {
				writeFile(zip, child);
			}
		}
	}

	private void writeFile(final ZipOutputStream zip, final IAggregationEntry child) throws IOException {
		writeMetadata(zip, child);
		final ZipEntry zipEntry = new ZipEntry(policy.getFilename(child, false).toString());
		zipEntry.setTime(child.getMetadata().getGeneric().getGenerated().getLastModified().toGregorianCalendar().getTimeInMillis());

		final ImportObject importObject = mapping.getImportObjectForTextGridURI(child.getTextGridURI().toString());
		try {
			final InputStream content = repository.getContent(child.getTextGridURI(), sid);
			if (importObject.getRewriteMethod().equals(RewriteMethod.XML)) {
				final ConfigurableXMLRewriter rewriter = new ConfigurableXMLRewriter(mapping, true);
				rewriter.configure(URI.create(importObject.getRewriteConfig()));
				final Optional<URI> base = policy.getBase(child);
				if (base.isPresent()) {
					rewriter.setBase(base.get());
				}
				final FileBackedOutputStream buffer = new FileBackedOutputStream(1024 * 1024);
				try {
					rewriter.rewrite(content, buffer);
					zip.putNextEntry(zipEntry);
					ByteStreams.copy(buffer.getSupplier(), zip);
				} catch (final XMLStreamException e) {
					final String errorMsg = MessageFormat.format("Failed to rewrite {0} (error: {1}). Exported with verbatim links instead.", child, e.getMessage());
					logger.log(Level.WARNING, errorMsg, e);
					zipEntry.setComment(errorMsg);
					importObject.setRewriteMethod(RewriteMethod.NONE);
					zip.putNextEntry(zipEntry);
					ByteStreams.copy(buffer.getSupplier(), zip);
				}
			}
			else {
				zip.putNextEntry(zipEntry);
				ByteStreams.copy(content, zip);
			}
			zip.closeEntry();
		} catch (final Exception e) {
			Throwables.propagateIfPossible(e, IOException.class);
			throw new WebApplicationException(e);
		}
	}

	private void writeMetadata(final ZipOutputStream zip, final IAggregationEntry child) throws IOException {
		final ZipEntry zipEntry = new ZipEntry(metaPolicy.getFilename(child, false).toString());
		zipEntry.setTime(child.getMetadata().getGeneric().getGenerated().getLastModified().toGregorianCalendar().getTimeInMillis());
		zip.putNextEntry(zipEntry);
		final ConfigurableXMLRewriter rewriter = new ConfigurableXMLRewriter(mapping, true);
		rewriter.configure(URI.create("internal:textgrid#metadata"));
		final Optional<URI> base = metaPolicy.getBase(child);
		if (base.isPresent()) {
			rewriter.setBase(base.get());
		}

		final FileBackedOutputStream buffer = new FileBackedOutputStream(1024*1024);
		JAXB.marshal(child.getMetadata(), buffer);
		try {
			rewriter.rewrite(buffer.getSupplier().getInput(), zip);
		} catch (final XMLStreamException e) {
			logger.log(Level.SEVERE, MessageFormat.format("Error rewriting the metadata of {0}. Should not happen.", child), e);
		}
		zip.closeEntry();
	}

}
